/* Copyright 2002 Ben Blum, Christian Shelton
 *
 * This file is part of GameTracer.
 *
 * GameTracer is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * GameTracer is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with GameTracer; if not, write to the Free Software Foundation, 
 * Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "gnmgame.h"
#include "cmatrix.h"
#include "math.h"

gnmgame::gnmgame(int numPlayers, int *actions): numPlayers(numPlayers) {
  int i;
  numStrategies = 1;
  for(i = 0; i < numPlayers; i++) {
    numStrategies *= actions[i];  // # of pure strat profiles
  }

  this->actions = new int[numPlayers];
  strategyOffset = new int[numPlayers+1];
  numActions = 0;
  maxActions = 0;
  for(i = 0; i < numPlayers; i++) {
    strategyOffset[i] = numActions;
    this->actions[i] = actions[i];
    numActions+=actions[i];
    if(actions[i] > maxActions)
      maxActions = actions[i];
  }
  strategyOffset[i] = numActions;
}
  
gnmgame::~gnmgame() {
  delete[] strategyOffset;
  delete[] actions;
}



void gnmgame::retractJac(cmatrix &dest, int *support) {
  int n, i, j;  
  Number totalk;
  for(n = 0; n < numPlayers; n++) {
    totalk = 0.0;
    for(i = firstAction(n); i < lastAction(n); i++) {
      totalk += support[i];
    }
    for(i = 0; i < numActions; i++) {
      for(j = firstAction(n); j < lastAction(n); j++) {
	if(i >= firstAction(n) && i < lastAction(n) && support[i] && support[j]) {
	  if(i==j)
	    dest[i][j] = ((Number)1.0) - (Number(1.0))/totalk;
	  else
	    dest[i][j] = ((Number)-1.0)/totalk;
	} else
	  dest[i][j] = 0.0;
      }
    }
  }
}

int compareDouble(const void *d1, const void *d2) {
  if(*(Number *)d1 > *(Number *)d2)
    return -1;
  else if(*(Number *)d1 < *(Number *)d2)
    return 1;
  else
    return 0;
}

void gnmgame::retract(cvector &dest, cvector &z) {//z: mixed strategy profile
  int n, i;
  Number v, sumz;
  Number y[numActions];
  CPY(y,z.values(),numActions);
  for(n = 0; n < numPlayers; n++) {  
    qsort(y+firstAction(n),actions[n],sizeof(Number),compareDouble);
    sumz = y[firstAction(n)];
    for(i=firstAction(n)+1; i < lastAction(n); i++) {
      if(sumz - ((Number)(i-firstAction(n))) * y[i] > (Number)1)
	break;
      sumz += y[i];
    }
    v = (sumz - (Number)1) / (Number)(i-firstAction(n));
    for(i = firstAction(n); i < lastAction(n); i++) {
      dest[i] = z[i] - v;
      if(dest[i] < (Number)0.0)
	dest[i] = 0.0;
    }
  }
}

Number gnmgame::LNM(cvector &z, const cvector &g, Number det, cmatrix &J, cmatrix &DG, cvector &s, int MaxLNM, Number fuzz, cvector &del, cvector &scratch, cvector &backup) {
  Number b, e = BIGFLOAT, ee;
  int k, faulted = 0;
  if(MaxLNM >= 1 && det != (Number)0.0) {
    b = ((Number)1.0)/det;
    for(k = 0; k < MaxLNM; k++) {
      //      del = z - s - DG*s / (Number)(numPlayers - 1) - g; 
      DG.multiply(s,del);
      del /= (Number)(numPlayers - 1);
      del += g;
      del += s;
      del -= z;
      del.negate();
      ee = max(del.max(),-del.min());

      if(ee < fuzz) {
	e = ee;
	break;
      } else if(e < ee) { // we got worse
	z = backup;
	retract(s, z);
	payoffMatrix(DG, s, fuzz);
      	if(faulted) // if we've already failed once, quit.
	  return e;
	b /= MaxLNM; // if the full LNM step fails to improve things,
	e = BIGFLOAT; // take smaller steps.
	faulted++;
	continue;
      }
      e = ee;
      J.multiply(del, scratch);
      scratch *= b;
      backup = z;
      z -= scratch;
      //      z = z - (J * del) * b;
      retract(s, z);
      payoffMatrix(DG, s, fuzz);
    }
    return ee;
  } else return fuzz;
}

void gnmgame::normalizeStrategy(cvector &s) {
  Number sum;  
  for(int n = 0; n < numPlayers; n++) {
    sum = 0.0;
    for(int i = firstAction(n); i < lastAction(n); i++) {
      sum += s[i];
    }
    for(int i = firstAction(n); i < lastAction(n); i++) {
      s[i] /= sum;
    }
  }
}

int indexOf(int *list, int target, int length) {
  for(int i = 0; i < length; i++) {
    if(list[i] == target)
      return i;
  }
  return -1;
}

void gnmgame::LemkeHowson(cvector &dest, cmatrix &T, int *Im) {
  Number D = 1;
  int cg = numActions + numPlayers ;
  int K = cg+1;
  int n, pc, pr, p;
  Number m;
  int col[numActions+numPlayers+2], row[numActions+numPlayers];
  for(n = 0; n < numActions+numPlayers+2; n++)
    col[n] = n+1;
  for(n = 0; n < numActions+numPlayers; n++)
    row[n] = -n-1;
  for(n = 0; n < numPlayers; n++) {
    pc = indexOf(col, Im[n]+1, numActions+numPlayers+2);
    pr = indexOf(row,-numActions-n-1, numActions+numPlayers);
    p = Pivot(T, pr, pc, row, col, D);
    pc = indexOf(col, numActions+n+1, numActions+numPlayers+2);
    pr = indexOf(row, -Im[n]-1, numActions+numPlayers);
    p = Pivot(T, pr, pc, row, col, D);
  }
  pc = indexOf(col, cg+1, numActions+numPlayers+2);
  m = -BIGFLOAT;
  pr = -1;
  for(n = 0; n < numPlayers + numActions; n++) {
    if(T[n][pc] < (Number)0) {
      if(T[n][K] / T[n][pc] > m) {
        m = T[n][K] / T[n][pc];
	pr = n;
      }
    }
  }

  if(m > (Number)0) {
    p = Pivot(T, pr, pc, row, col, D);
    do {
      pc = indexOf(col, -p, numActions+numPlayers+2);
      m = BIGFLOAT;
      pr = -1;
      for(n = 0; n < numPlayers + numActions; n++) {
	if(T[n][pc] > (Number)0 && (row[n] <= numActions || row[n] > numActions+numPlayers)) {
	  if(T[n][K] / T[n][pc] < m) {
	    m = T[n][K] / T[n][pc];
	    pr = n;
	  }
	}
      }
      p = Pivot(T, pr, pc, row, col, D);
    } while(p != cg+1);
  }
  for(n = 0; n < numActions; n++) {
    pr = indexOf(row,n+1,numActions+numPlayers);
    if(pr == -1)
      dest[n] = 0.0;
    else
      dest[n] = T[pr][K] / D;
  }
}

int gnmgame::Pivot(cmatrix &T, int pr, int pc, int *row, int *col, Number &D) {
  Number pivot = T[pr][pc];
  int i0,j0,p,sgn = pivot < (Number)0 ? -1 : 1;
  
  for(i0 = 0; i0 < numActions+numPlayers; i0++) {
    if(i0 != pr) {
      for(j0 = 0; j0 < numActions+numPlayers+2; j0++) {
	if(j0 != pc) {
	  T[i0][j0] *= pivot;
	  T[i0][j0] -= T[i0][pc] * T[pr][j0];
	  T[i0][j0] /= D * (Number)sgn;
	}
      }
    }
  }
  if(sgn == 1) {
    for(i0 = 0; i0 < numActions+numPlayers; i0++) {
      T[i0][pc] = -T[i0][pc];
    }
  } else {
    for(j0 = 0; j0 < numActions+numPlayers+2; j0++) {
      T[pr][j0] = -T[pr][j0];
    }
  }
  T[pr][pc] = ((Number)sgn) * D;
  D = pivot < (Number)0 ? -pivot : pivot;
  p = row[pr];
  row[pr] = col[pc];
  col[pc] = p;
  if(D > (Number)1e6) {
    T /= D;
    D = 1;
  }
  return p;
}

   
