/* Copyright 2002 Ben Blum, Christian Shelton
 *
 * This file is part of GameTracer.
 *
 * GameTracer is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * GameTracer is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with GameTracer; if not, write to the Free Software Foundation, 
 * Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef __GNMGAME_H
#define __GNMGAME_H

#include "cmatrix.h"
#define BIGFLOAT 3.0e+28F

#ifdef WIN32
#ifndef drand48
#define srand48(x) srand(x)
#define drand48(x) ((double)rand(x)/RAND_MAX)
#endif
#endif

class gnmgame {
 public:
  
  //actions[i] = number of actions player i has
  gnmgame(int numplayers, int *actions);
  virtual ~gnmgame();
  
  
  
  // Input: s[i] has integer index of player i's pure strategy
  // s is of length numPlayers

  virtual Number getPurePayoff(int player, int *s) = 0; 

  //commented out by albert

  //virtual inline void setPurePayoff(int player, int *s, Number value) = 0;

  // The actions of all players are combined in one linear array of length
  // numActions; this gives the index of a player's first action in the array.
  inline int firstAction(int player) {
    return strategyOffset[player];
  }

  inline int lastAction(int player) {
    return strategyOffset[player+1];
  }

  // s is the mixed strategy profile.  It is of length numActions, and
  // s[i] is the probability that the appropriate player takes action i.
  virtual Number getMixedPayoff(int player, cvector &s) = 0;

  // s is the mixed strategy profile, as above.  This function stores
  // the Jacobian of the payoff function G, where G(i) is the payoff to
  // the owner of action i if he deviates from s by choosing i instead.
  virtual void payoffMatrix(cmatrix &dest, cvector &s, Number fuzz) = 0;

  // store in dest the the payoff function G, where G(i) is the payoff to
  // the owner of action i if he deviates from s by choosing i instead.
  virtual void getPayoffVector(cvector &dest, int player,const cvector &s) = 0; 

  //get payoff vector for a symmetric game under a symmetric strategy. only one player's strategy is given in s
  virtual void getSymPayoffVector(cvector &dest, cvector &s){
    cvector fulls(getNumActions());
    int nact=getNumActions(0);
    for(int i=0;i<getNumPlayers();++i){
      assert(nact==getNumActions(i));
      for(int j=0;j<nact;++j)
        fulls[j+firstAction(i)]=s[j];
    }
    getPayoffVector(dest, 0, fulls);
  }

  // return the regret for a player under strategy profile s
  Number getRegret(int player, cvector &s)
  {
    Number p=0;
    cvector payoffs(actions[player]);
    getPayoffVector(payoffs, player, s);
    for (int i=0;i<actions[player];i++){
      p+=payoffs[i] * s[firstAction(player)+i];
    }
    return payoffs.max()-p;
  }

  //return the max regret under strategy profile s
  Number getRegret(cvector &s)
  {
    cvector regrets(numPlayers);
    for (int i=0;i<numPlayers;i++) regrets[i]= getRegret(i,s);
    return regrets.max();
  }


  // this stores the Jacobian of the retraction function in dest.  
  void retractJac(cmatrix &dest, int *support);

  // This retracts z onto the nearest normalized strategy profile, according
  // to the Euclidean metric
  void retract(cvector &dest, cvector &z);

  // LNM runs the local Newton method on z to attempt to bring it closer to
  // the image of the graph of the equilibrium correspondence above the ray,
  // under the homeomorphism.  In order to prevent costly memory allocation,
  // a number of scratch vectors are passed in.

  Number LNM(cvector &z, const cvector &g, Number det, cmatrix &J, cmatrix &DG,  cvector &s, int MaxLNM, Number fuzz, cvector &del, cvector &scratch, cvector &backup);

  // This normalizes a strategy profile by scaling appropriately.
  void normalizeStrategy(cvector &s);

  void LemkeHowson(cvector &dest, cmatrix &T, int *Im);


  inline int getNumPlayers() { return numPlayers; }
  inline int getNumActions() { return numActions; }
  inline int getNumActions(int p) { return actions[p]; }
  inline int getMaxActions() { return maxActions; }

  //albert
  // generate random strat profile, with full support
  void randomFullStrategy(cvector& dest, int player){
      Number normconst;
      normconst=0;
      for (int j=firstAction(player); j<lastAction(player); j++){
            dest[j] = drand48();
            normconst+= dest[j];
      }
      for (int j= firstAction(player);j<lastAction(player); j++)
            dest[j] /= normconst;
            
  }
  void randomFullStrategy (cvector& dest){
    for (int i=0; i<numPlayers; ++i){
      randomFullStrategy(dest, i);
    }
  }
  // generate random strat profile, with random support size.
  // returns the volume of the support profile.
  unsigned long long randomSupportStrategy(cvector& dest,int player, Number posprob){
      Number normconst;
      unsigned long long currsupp;
      
      do{
        currsupp=0;
        normconst=0;
        for (int j=firstAction(player);j<lastAction(player); j++){
          if ( drand48() < posprob) {
            dest[j] = drand48();
            normconst+=dest[j];
            currsupp++;
          }
          else {
            dest[j]=0;
          }
        }
      } while (currsupp<=0 || normconst==(Number)0.0);
      
      // normalize
      for (int j=firstAction(player);j<lastAction(player); j++)
        dest[j] /= normconst;
        
      return currsupp;
  }

  unsigned long long randomSupportStrategy(cvector& dest , Number posprob){
    unsigned long long suppsize=1;
    for (int i=0;i<numPlayers; ++i){
      suppsize *= randomSupportStrategy(dest, i, posprob);
    }
    return suppsize;
  }

 protected:
  
  int Pivot(cmatrix &T, int pr, int pc, int *row, int *col, Number &D);

  int *strategyOffset;
  int numPlayers, numStrategies, numActions;
  int *actions;
  int maxActions;
};

#endif
