
#ifndef __TRIE_MAP_H
#define __TRIE_MAP_H

//Trie with STL-like interfaces.
//Mapping from vector of ints to type V.
//WARNING: traversal using the iterators is in the reverse order of insertion.

#include <math.h>
#include <ext/slist>
#include <iterator>
using namespace std;
using __gnu_cxx::slist;

//forward declarations

template <class V>
class trie_map;

template <class V>
ostream& operator<< (ostream& s, const trie_map<V>& t);



template <class V>
struct TrieNode
{
  //members:
  vector<TrieNode*> children;
  typename slist<pair<vector<int>,V> >::iterator val;


  //methods:

  //constructor  
  TrieNode(size_t branches, typename slist<pair<vector<int>,V> >::iterator v):children(branches,(TrieNode*)NULL),val(v) {}
  

  inline TrieNode*& operator[](size_t i){
    if (i<children.size()) return children[i];
    else {children.resize(i+1,(TrieNode*)NULL); return children[i];}
  }
};

template <class V>
class trie_map {

public:
  //typedefs
  typedef vector<int>          key_type;
  typedef pair<vector<int>, V> value_type;

  typedef V* 		pointer;
  typedef V&		reference;

  typedef unsigned int  size_type;

  typedef typename slist<value_type>::iterator iterator;
  typedef typename slist<value_type>::const_iterator const_iterator;


  //friends
  friend ostream& operator<< <V>(ostream& s, const trie_map<V>& t);
  
  friend class agg;

  //constructors
  trie_map():initBranches(1),root(new TrieNode<V>(1,data.end())){}
  trie_map(int branches):initBranches(branches),root(new TrieNode<V>(branches,data.end())){}


  //copy constructor
  trie_map(const trie_map<V>& other);

  //destructor
  virtual ~trie_map(){deleteNodes(root);}

  //assignment
  inline trie_map<V>& operator=( const trie_map<V>& other);

  inline void swap (trie_map<V>& other);

  inline size_type size() const {return data.size();}
  inline bool empty() const {return data.empty();}
  inline iterator begin() {return data.begin();}
  inline iterator end() {return data.end();}
  inline const_iterator begin() const {return data.begin();}
  inline const_iterator end() const {return data.end();}

  //insert: same interface as in STL map
  inline pair<iterator,bool> insert (const value_type& x);

  template <class InputIterator>
  inline void insert(InputIterator f, InputIterator l){
    for (InputIterator p=f; p != l; ++p){
	insert( *p);
    }
  }

  //insert or add
  inline trie_map<V>& operator+=(const value_type& x){
    pair<typename trie_map<V>::iterator ,bool> r=insert(x);
    if (!r.second) (*r.first).second += x.second;
    return (*this);
  }


  //prefix matching
  //return a reference to iterator
  inline iterator& find (const key_type& k) __attribute__((always_inline));

  //exact matching
  inline iterator findExact (const key_type& k);

  //clear the tree strucutre as well as data
  inline void clear(){
	deleteNodes(root);
	root=new TrieNode<V> (initBranches, data.end()); 
	data.clear();
	leaves.clear();
  }

  //clear data, but keep the tree structure
  inline void reset(){
    while(!leaves.empty()) {

#ifdef AGGDEBUG
      if(!leaves.back()){
	  cout<<"WARNING: NULL pointer in leaves. leaves are:"<<endl;
	  for(size_t i=0;i<leaves.size();++i){ 
	      cout << (int) leaves[i]<<" ";
	      
	      if (leaves[i])cout<<leaves[i]->val->second  <<endl;
	  }
	  cout<<endl<<"trie is:"<< *this<< endl;
	  cout<<"in order:"<<endl;
	  print_in_order();
      }
#endif


      leaves.back()->val=data.end();
      leaves.pop_back();
    }
    data.clear();
  }

  //number of elements with key exactly k.
  // returns 1 or 0 
  inline size_type count(const key_type& k) {
    size_t i=0;
    TrieNode<V>* ptr=root;
    for (;i<k.size()&&k[i]<(int)ptr->children.size()&&  ptr->children[k[i]];ptr=ptr->children[k[i++]]) ;
    
    return (i>=k.size()&&ptr->val!=data.end());
  }

  //in order traversal
  template <class UnaryFunction>
  inline void in_order(UnaryFunction f, bool debug=false){
	in_order_subtree(f, root, debug);
  }
  inline void print_in_order (){in_order(print_helper(end()),true);}

  //polynomial multiplication
  void multiply (const trie_map<V>& t1,const trie_map<V>& t2,size_t keylen,
	 vector<proj_func*>& f)
  {
    static pair<vector<int>, V> v;
    v.first.resize(keylen);
    reset();
    for (const_iterator p1=t1.begin(); p1!=t1.end(); ++p1)if((*p1).second>(V)0){
      for(const_iterator p2=t2.begin(); p2!=t2.end(); ++p2)if((*p2).second>(V)0){
	assert((*p1).first.size()==keylen&& (*p2).first.size()==keylen);
	//if ( ){
	//  transform((*p1).first.begin(),(*p1).first.end(),
	//	(*p2).first.begin(), k.begin(), plus<int>() );
	//}else
	for (size_t i=0;i<keylen;++i){
	  v.first[i]= (*(f[i])) ((*p1).first[i], (*p2).first[i]);
	}
	v.second = (V)((*p1).second * (*p2).second);
	(*this) += v;
      }//end for(p2
    }//end for(p1 
  }
  //multiply in-place
  void multiply (const trie_map<V>& other,size_t keylen, vector<proj_func*>& f);

  //inner product
  V inner_prod( trie_map<V>& other, V init= (V)(0) ) const{
    V result(init);
    //V th(THRESH);
    for(const_iterator p=begin(); p!=end(); ++p)if((*p).second>(V)0){
	typename trie_map<V>::iterator p2=other.find( (*p).first);
	//assert(p2 != other.end());
	if (p2==other.end()){
	  cout<<"inner_prod WARNING: discarding [";
	  copy(p->first.begin(),p->first.end(), ostream_iterator<int>(cout," "));
	  cout<<"] "<<p->second<<endl;
	} 
	else {
	  result+= (*p).second * (*p2).second;
	}
    }
    return result;
  }

  //first apply the action x, then inner prod
  V inner_prod(const vector<int>& x, size_t keylen, vector<proj_func*>& f,
	trie_map<V>& other, V init=(V)(0) ) const
  { 
    V result(init);
    //V th(THRESH);
    static iterator p2;
    //V s(-1);
    for (const_iterator p=begin(); p!=end();++p)if((*p).second>(V)0){
      value_type y= *p;
      //assert(y.first.size()==keylen);
      for (size_t i=0; i<keylen;++i){
	y.first[i] = (*(f[i])) (y.first[i],x[i]); 
      }
      //s += y.second;

      p2 = other.find(y.first);
      //assert(p2!= other.end());
      if (p2==other.end()){
	cout<<"inner_prod WARNING: discarding [";
	copy(y.first.begin(),y.first.end(), ostream_iterator<int>(cout," "));
	cout<<"] "<<y.second<<endl;
      } else{
        result += y.second  * (*p2).second;
      }
    }

//    if (s>THRESH ||s<-THRESH){
//	cout<<"inner_prod WARNING: sum differ from 1.0: diff is"<<s<<endl;
//    } 
    return result;
  }

  
  //polynomial division
  inline trie_map<V>& operator/= (const vector<V>& denom);

  
private:
  //member variables:
  slist<value_type> data;
  size_type initBranches; //default branching factor
  TrieNode<V> *root;

  vector<TrieNode<V>*> leaves;

  static const double  THRESH = 1e-12;
  
  
  struct print_helper {
   print_helper (typename trie_map<V>::iterator en):endp(en){}
   void operator()(typename trie_map<V>::iterator p){
    if (p== endp) {cout<< "leaf with no data"<<endl; return;}
    cout<<"[";
    copy(p->first.begin(),p->first.end(),ostream_iterator<int>(cout," "));
    cout <<"]: "<<p->second<<endl;
   }
   typename trie_map<V>::iterator endp;
  };



  //private methods:
  //div: helper for operator/=()
  void  div(const vector<V>& denom, TrieNode<V>* n, int current, int pivot);



  __attribute__((noinline)) void deleteNodes(TrieNode<V> *n){
	if (!n) return;
	size_type i,s=n->children.size();
	for (i=0;i<s;++i) if (n->children[i]) {
	  deleteNodes (n->children[i]);
	}
	delete n;
  }


  /*
  __attribute__((noinline)) void remove (TrieNode<V> *n){
	if (!n) return;
	size_type i, s=n->children.size();
	for (i=0;i<s;++i) if (n->children[i]){
	  remove(n->children[i]);
	}

	//if (n->val!=data.end()) {
        //    n->val->second=0;
        //    n->val= data.end();
        //}
	//else {
        //    delete n;
        //}
  }
  */


  //in order traversal
  template <class UnaryFunction>
  __attribute__((noinline)) void in_order_subtree(UnaryFunction& f, TrieNode<V> *n, bool debug=false)
  {
    if (n==NULL) return;
    size_type i,s=n->children.size();
    bool is_leaf=true;
    for (i=0;i<s;++i)if (n->children[i]){
	is_leaf=false;
	if (debug) cout<< i<<" ";
	in_order_subtree(f, n->children[i], debug);
    }
    if(is_leaf) {
	if(debug) cout<< "(leaf): ";
	f(n->val);
    }
    return;
  }    



  //private helper functor classes:

  struct div_helper :public unary_function<void, iterator>{
    div_helper(const vector<V>& den, int piv,iterator en)
	:pivot(piv), denom(den),endp(en){}
    inline void operator()(iterator p){
	if (p==endp) return;
	p->first[pivot]--;
	p->second /= denom[pivot];
    } 
    int pivot;
    const vector<V>& denom;
    iterator endp;
  };

  struct div_helper_mul: public unary_function<void,iterator>{
    div_helper_mul(const vector<V>& den, int piv, TrieNode<V>* des,iterator en)
	:pivot(piv),denom(den), dest(des),endp(en) {}

    void add(const vector<int>& conf, V y){
      size_t i,keylen=conf.size();
      V th(THRESH);
      if (y<=th&&y>=-th) return;

      if (!dest){
	cout<<"division (pivot=" << denom[pivot]<<") WARNING: discarding "
	    <<y<<endl;
        return;
      }
      TrieNode<V>* ptr=dest;
      for (i=pivot+1; i<keylen; ++i){
	if( (*ptr)[conf[i]]==NULL){
	    cout<<"division (pivot=" << denom[pivot]<<") WARNING: discarding "
	    <<y<<endl;
	    return;
	}
	ptr = (*ptr)[conf[i]];
      }
      ptr->val->second += y;
      //assert(ptr->val->second>-THRESH);
      if (ptr->val->second <= -th)
	cout<<"division (pivot=" << denom[pivot]<<") WARNING: discarding "
	    <<ptr->val->second<<endl;
      if (ptr->val->second <(V)0) ptr->val->second=0;
    }
    void operator()(iterator p){
      if (p==endp) return;
      size_t i, keylen = p->first.size();
      assert(keylen==denom.size());
      V null_prob(((V)1)-denom[pivot]);
      V th(THRESH);
      for (i=pivot+1; i<keylen; ++i)if(denom[i]>th){
	p->first[i]++;
	add(p->first, -denom[i]* p->second);
	p->first[i]--;
	null_prob-= denom[i];
      }
      
      if (null_prob>th)add(p->first, -null_prob * p->second);
    }

    
    int pivot;
    const vector<V>& denom;
    TrieNode<V>* dest;
    iterator endp;
  };  //end struct div_helper_mul

};

template <class V>
inline ostream& operator<< (ostream& s, const trie_map<V>& t)
{
  //s<<endl;
  for (typename trie_map<V>::const_iterator p=t.begin();p!=t.end();++p){
    s<<"[ ";
    copy(p->first.begin(),p->first.end(),ostream_iterator<int>(s, " "));
    s<< "] "<<p->second<<endl;
  }
  return s;
}



#include "trie_map.template"
#endif
