/* Copyright 2002 Ben Blum, Christian Shelton
 *
 * This file is part of GameTracer.
 *
 * GameTracer is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * GameTracer is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with GameTracer; if not, write to the Free Software Foundation, 
 * Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include<fstream>
#include "nfgame.h"
#include "libagg/proj_func.h"
#include "libagg/agg.h"
#include "BAGG/bagg.h"
#include "makegame.h"

nfgame *makeRandomNFGame(int n, int actions, int seed) {
  int sizes[n];
  srand48(seed);
  int total = n;
  for(int i = 0; i < n; i++) {
    sizes[i] = actions;
    total *= actions;
  }
  cvector payoffs(total);
  for(int i = 0; i < total; i++) {
    payoffs[i] = drand48();
  }
  return new nfgame(n,sizes,payoffs);
}

nfgame *makeNFGame(char *filename) {
  int i, n, total, actions;
  ifstream in(filename);
  cout<<"Reading a normal form game from "<<filename<<endl;

  /* Now read in a game file.  The format for a game file is:
     numPlayers
     numActions[0] numActions[1] ... numActions[numPlayers-1]
     payoff(0,0,...,0) payoff(1,0,...,0) ... payoff(numActions[0]-1,numActions[1]-1,...,numActions[numPlayers-1]-1)
  */
  if(in.good() && !in.eof()) {
    in >> n;
    if (!in) {
      cout<<"Error in reading game file: number of players expected."<<endl;
      return 0;
    }
    int size[n];
    total = n;
    actions = 0;
    for(i = 0; i < n; i++) {
      if(in.eof()) {
	cout << "Error in game file: not enough actions.\n";
	return 0;
      }
      in >> size[i];
      if(!in){
        cout<<"Error in reading game file: expected an integer for number of player "<<i<<"\'s actions."<<endl;
        return 0;
      }
      total *= size[i];
      actions += size[i];
    }
    if(total <= 0) {
      cout << "Error in game file: illegal actions.\n";
      return 0;
    }
    cout<<"Making an Normal Form game with "<<total<<"entries"<<endl;
      
    cvector payoffs(total);
    for(i = 0; i < total; i++) {
      if(in.eof()) {
	cout << "Error in game file: not enough payoffs.\n";
	return 0;
      }
      in >> payoffs[i];
      if(!in){
        cout<<"Error in game file: number expected.";
        return 0;
      }
    }	
    return new nfgame(n, size, payoffs);
  } else {
    cout << "Bad game file.\n";
    return 0;
  }
}

int  getPureStrat (int *s , int index,int* blockSize,int numPlayers){
    int player = index / blockSize[numPlayers];
    index %= blockSize[numPlayers];
    for (int i=numPlayers-1;i>=0; i--){
      s[i] = index / blockSize[i];
      index %= blockSize[i];
    }
    return player;
}

//albert
nfgame *makeNFGame(gnmgame *aggPtr){
  
  int n= aggPtr->getNumPlayers();
  unsigned long long total=n;
  int sizes[n];
  int s[n];
  for (int i=0; i<n; i++) {
    sizes[i]=aggPtr->getNumActions(i);
    total *= sizes[i];
  }
  cout<<"makeNFGame(): making a game with "<<total<< " payoff entries"<<endl;
  cvector payoffs (total); 
  int*  blockSize = new int[n + 1];
  blockSize[0] = 1;
  for(int i = 1; i <= n; i++) {
    blockSize[i] = blockSize[i-1]*sizes[i-1];
  }
  
  for (unsigned long long j=0; j<total; j++){
    int player = getPureStrat(s,j,blockSize,n);
    payoffs[j]= aggPtr->getPurePayoff(player,s);
  }
  delete [] blockSize;
  cout<<"makeNFGame(): finished converting the payoffs"<<endl;
  return new nfgame(n,sizes,payoffs);
}

nfgame* makeNFGame(bagg *baggPtr)
{
  int n= baggPtr->getNumPlayers();
  unsigned long long total=n;
  int sizes[n];
  //int s[n];
  int s[baggPtr->getNumTypes()];
  for (int i=0; i<n; i++) {
    sizes[i]=1;
    for (int j=0;j<baggPtr->getNumTypes(i);j++){    
      sizes[i]*=baggPtr->getNumActions(i,j);
    }
    total *= sizes[i];
  }
  cout<<"makeNFGame(): making a game with "<<total<< " payoff entries"<<endl;
  cvector payoffs (total); 
  int*  blockSize = new int[baggPtr->getNumTypes() + 1];
  blockSize[0] = 1;
  int idx=1;
  for(int i = 0; i < n; i++) {
    for(int j=0;j<baggPtr->getNumTypes(i);++j, ++idx){
      blockSize[idx] = blockSize[idx-1]*baggPtr->getNumActions(i,j);
    }
  }
  
  for (unsigned long long j=0; j<total; j++){
    int player = getPureStrat(s,j,blockSize,baggPtr->getNumTypes());
    payoffs[j]= baggPtr->getPurePayoff(player,s);
  }
  delete [] blockSize;
  cout<<"makeNFGame(): finished converting the payoffs"<<endl;
  return new nfgame(n,sizes,payoffs);

}

agg* makeRandSymAGG(int n, int S, int seed){
  int P=0;
  int sizes[n];
  srand48(seed);

  for(int i = 0; i < n; i++) {
    sizes[i] = S;

  }
  vector<projtype> projTypes (P);
  vector <vector<int> >ASets (n);
  for (int i=0;i<n;i++){
    for (int j=0;j<S; j++){
      ASets[i].push_back(j);
    }
  }
  vector<vector<int> >neighb (S);
  for(int i=0;i<S;i++) 
    for (int j=0; j<S; j++)
      neighb[i].push_back (j);
  
  
  return agg::makeRandomAGG(n,sizes, S,P,ASets, neighb, projTypes,seed );
}

agg* makeRandCoffeeShopAGG(int n, int r,int c,int seed,projtype pType, 
bool full){
  int S = r*c+1;
  int P = r*c;
  if (full) P+=r*c;
  int sizes[n];
  int dr[8] ={-1,-1,-1,0,0,1,1,1}; 
  int dc[8] ={-1,0,1,-1,1,-1,0,1};
  srand48(seed);
  for (int i=0;i<n;++i) sizes[i]=S;
  
  vector<vector<int> >ASets(n);
  for (int i=0;i<n;i++){
    for (int j=0;j<S; j++){
      ASets[i].push_back(j);
    }
  }  
  vector<vector<int> >neighb (S+P);
  neighb[0].clear();
  for (int i=1;i<S;i++){
    neighb[i].push_back(i);
    neighb[i].push_back(i + r*c);
    if (full)neighb[i].push_back(i+2*r*c); 
  }
  for (int i=0; i<r; i++){
    for (int j=0;j<c;j++){
      int node = i*c+j +S;
      for (int k=0;k<8;k++){
        int i1=i+dr[k];
        int j1= j+dc[k];
        if(i1<0||i1>=r || j1<0||j1>=c) continue;
        neighb[node].push_back(i1*c+j1+1);
      }
    }
  }
  
  if (full)
   for (int i=0; i<r; i++)
    for (int j=0;j<c;j++)
      for (int i1=0;i1<r;i1++)
        for (int j1=0;j1<c;j1++){
          int node = i*c+j + S+r*c;
          if (abs (i-i1) >1 ||abs(j-j1) >1)
            neighb.at(node).push_back(i1*c+j1+1);
        
        }  

  vector<projtype> projTypes (P, pType);
  return agg::makeRandomAGG(n,sizes, S,P,ASets, neighb, projTypes,seed );  
}


bagg* makeRandCoffeeShopBAGG(int n,int T, int r,int c,int seed,projtype pType, bool int_payoffs, int int_factor){
  int S = T*r*c+1;
  int P = 3*r*c;
  vector<int> numTypes(n,T);
  vector<cvector> TDist(n, cvector(T,1.0/T));
  int dr[8] ={-1,-1,-1,0,0,1,1,1}; 
  int dc[8] ={-1,0,1,-1,1,-1,0,1};
  srand48(seed);
    
  
  vector<vector<vector<int> > > TASets(n);
  for (int i=0;i<n;i++){
    for (int j=0;j<T; j++){
      TASets[i].push_back(vector<int>());
      TASets[i][j].push_back(0);
      for(int k=0;k<r*c;k++){
        TASets[i][j].push_back(j*r*c + k + 1);
      }
    }
  }  
  vector<vector<int> > neighb (S+P);
  neighb[0].clear();

  //edges into action nodes
  for (int i=0;i<T;i++){
    for(int j=0;j<r*c;j++){
      neighb[i*r*c+j+1].push_back(S+j);
      neighb[i*r*c+j+1].push_back(S+r*c+j);
      neighb[i*r*c+j+1].push_back(S+2*r*c+j);
    } 
  }

  for (int i=0;i<r*c;i++){
    for (int tp = 0;tp<T;tp++){
      neighb[S+i].push_back(tp*r*c + i + 1);
    }
  }

  for (int i=0; i<r; i++){
    for (int j=0;j<c;j++){
      int node = i*c+j +S+r*c;
      for (int k=0;k<8;k++){
        int i1=i+dr[k];
        int j1= j+dc[k];
        if(i1<0||i1>=r || j1<0||j1>=c) continue;
        neighb[node].push_back(i1*c+j1+S);
      }
    }
  }
  
  for (int i=0; i<r; i++)
    for (int j=0;j<c;j++)
      for (int i1=0;i1<r;i1++)
        for (int j1=0;j1<c;j1++){
          int node = i*c+j + S+2*r*c;
          if (abs (i-i1) >1 ||abs(j-j1) >1)
            neighb.at(node).push_back(i1*c+j1+S);
        
        }  

  vector<projtype> projTypes (P, pType);
  return bagg::makeRandomBAGG(n,numTypes,TDist, S,P,TASets, neighb, projTypes,seed, int_payoffs,int_factor);  
}

