/* Copyright 2002 Ben Blum, Christian Shelton
 *
 * This file is part of GameTracer.
 *
 * GameTracer is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * GameTracer is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with GameTracer; if not, write to the Free Software Foundation, 
 * Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <vector>
#include "cmatrix.h"
#include "nfgame.h"



nfgame::nfgame(int numPlayers, int *actions, const cvector &payoffs) : gnmgame(numPlayers, actions), payoffs(payoffs) {
  blockSize = new int[numPlayers + 1];
  blockSize[0] = 1;
  for(int i = 1; i <= numPlayers; i++) {
    blockSize[i] = blockSize[i-1]*actions[i-1];
  }
}

nfgame::~nfgame() {
  delete[] blockSize;
}

int nfgame::findIndex(int player, int *s) {
  int i, retIndex=player * blockSize[numPlayers];
  for(i = 0; i < numPlayers; i++)
    retIndex += blockSize[i] * s[i];
  return retIndex;
}

Number nfgame::getMixedPayoff(int player, cvector &s) {
  Number *m=NULL;
  m = new Number[blockSize[numPlayers]+1];
  assert (m);
  CPY(m, payoffs.values() + player * blockSize[numPlayers], blockSize[numPlayers]);
  
  Number res =  localPayoff(s, m, numPlayers-1);
  delete [] m;
  return res;
}

void nfgame::getPayoffVector(cvector &dest, int player, const cvector &s){
  //Number *m=NULL;
  //m=new Number[blockSize[numPlayers]+1];
  //assert(m);
  //CPY(m, payoffs.values() + player * blockSize[numPlayers], blockSize[numPlayers]);
  //for (int i=0;i<blockSize[numPlayers];++i) m[i]=payoffs[player*blockSize[numPlayers]+i];

  vector<Number> t(payoffs.values()+player*blockSize[numPlayers], payoffs.values()+(player+1)*blockSize[numPlayers]);

  localPayoffVector(dest.values(), player,s,&(t[0]),numPlayers-1);

  //delete [] m;
}

void nfgame::payoffMatrix(cmatrix &dest, cvector &s, Number fuzz) {
  //compute the jacobian of payoff function, DG
  //s: mixed strat
  int rown, coln, rowi, coli;
  Number fuzzcount;
  //Number m[blockSize[numPlayers]];
  Number *m = NULL;
  m = new Number[blockSize[numPlayers]+1];
  assert (m);

  Number local[maxActions*maxActions];
  for(rown = 0; rown < numPlayers; rown++) {
    for(coln = 0; coln < numPlayers; coln++) {
      if(rown == coln) {
	fuzzcount = fuzz;
	for(rowi=firstAction(rown); rowi < lastAction(rown); rowi++) {
	  for(coli=firstAction(coln); coli < lastAction(coln); coli++) {
	    dest[rowi][coli]=fuzzcount;
	    fuzzcount += fuzz;
	  }
	}
      } else {
	// set m to be the payoffs for player rown
	CPY(m, payoffs.values() + rown * blockSize[numPlayers], blockSize[numPlayers]);
	localPayoffMatrix(local, rown, coln, s, m, numPlayers-1);
	for(rowi = firstAction(rown); rowi < lastAction(rown); rowi++) {
	  for(coli = firstAction(coln); coli < lastAction(coln); coli++) {
	    if(rown > coln) {
	      dest[rowi][coli] = *(local + (rowi - firstAction(rown))*actions[coln] + (coli - firstAction(coln)));
	    } else {
	      dest[rowi][coli] = *(local + (coli - firstAction(coln))*actions[rown] + (rowi - firstAction(rown)));
	    }
	  }
	}
      }
    }
  }

  delete [] m;
}


//assumes m = memcpy(m, payoffs + blockSize[numPlayers] * player1, blockSize[numPlayers]*sizeof(Number)), player1 != player2
//i.e. m points to payoff cmatrix for the desired player

void nfgame::localPayoffMatrix(Number *dest, int player1, int player2, cvector &s, Number *m, int n) {
  //does the payoff Jacobian for player1 and player2's actions
  //recursive on n, whose mixed strategy we're summing up now
  //dest: actions[rown(or player1)] by actions[coln(player2)]
  //         if rown>coln;
  //      the other way around if coln>rown 
  int i;
  if(player1 == n) {
    for(i = 0; i < actions[player1]; i++) {
      localPayoffVector(dest+i*actions[player2], player2, s, m+i*blockSize[player1], n-1);
    }
  } else if(player2 == n) {
    for(i = 0; i < actions[player2]; i++) {
      localPayoffVector(dest+i*actions[player1], player1, s, m+i*blockSize[player2], n-1);
    }
  } else {
    m = scaleMatrix(s, m, n);
    localPayoffMatrix(dest, player1, player2, s, m, n-1);
  }
}

Number *nfgame::scaleMatrix(const cvector &s, Number *m, int n) {
  //only m[newbase : newbase+blockSize[n]] are modified;
  //applied player n's mixed strategy
  int i,j, curbase, newbase = -1;
  Number scale;
  for(i = 0; i < actions[n]; i++) {
    if(s[i+firstAction(n)] > (Number)0) {
      scale = s[i+firstAction(n)];//mixed strat probability
      curbase = i*blockSize[n];
      if(newbase < 0) {
	newbase = curbase;
	for(j = curbase; j < curbase+blockSize[n]; j++) {
	  m[j] *= scale;
	}
      } else {
	for(j = 0; j < blockSize[n]; j++) {
	  m[newbase + j] += scale * m[curbase+j];
	}
      }
    }
  }
  return m+newbase;
}

void nfgame::localPayoffVector(Number *dest, int player,const cvector &s, Number *m, int n) {
  //does one row of the local payoff matrix(jacobian)
  //player: the other player
  //recursive on n
  if(player == n) {
    for(int i = 0; i < actions[player]; i++) {
      dest[i] = localPayoff(s, m+i*blockSize[player], n-1);
    }
  } else {
    m = scaleMatrix(s, m, n);
    localPayoffVector(dest, player, s, m, n-1);
  }
}

Number nfgame::localPayoff(const cvector &s, Number *m, int n) {
  //does one cell of the local payoff matrix(jacobian)
  if(n < 0)
    return *m;
  else {
    m = scaleMatrix(s, m, n);
    return localPayoff(s, m, n-1);
  }
}
